
XIncludeFile "Container.pb"


Interface TreeMapEntry
  getKey.i()
  getValue.i()
  setValue.i(*value)
EndInterface


Interface TreeMap
  size.i()
  isEmpty.i()
  put.i(*key, *value=#Null, *isAdd.Integer=#Null)
  putIfAbsent(*key, *value=#Null, *isAdd.Integer=#Null)
  getEntry.i(*key)
  getKey.i(*key)
  getValue.i(*key)
  getOrDefault.i(*key, *defValue)
  contains.i(*key)
  delete.i(*key)
  deleteEntry.i(*entry)
  reset.i()
  prevEntry.i(*pointer.Integer=#Null)
  nextEntry.i(*pointer.Integer=#Null)
  prevKey.i(*pointer.Integer=#Null)
  nextKey.i(*pointer.Integer=#Null)
  prevValue.i(*pointer.Integer=#Null)
  nextValue.i(*pointer.Integer=#Null)
  getCurrentEntry.i()
  setCurrentEntry.i(*entry)
  getCurrentKey.i()
  getCurrentValue.i()
  clear.i()
  free.i()
EndInterface


Structure TreeMapEntryBase
  *VTable
  *parent.TreeMapEntryBase
  *left.TreeMapEntryBase
  *right.TreeMapEntryBase
  color.i
EndStructure


#TreeMapColorBlack = 0
#TreeMapColorRed = 1


;- TreeMap_prGetParent
Procedure TreeMap_prGetParent(*entry.TreeMapEntryBase)
  If (*entry <> #Null)
    ProcedureReturn *entry\parent
  EndIf
  ProcedureReturn #Null
EndProcedure


;- TreeMap_prGetColor
Procedure TreeMap_prGetColor(*entry.TreeMapEntryBase)
  If (*entry <> #Null)
    ProcedureReturn *entry\color
  EndIf
  ProcedureReturn #TreeMapColorBlack
EndProcedure


;- TreeMap_prRotateLeft
Procedure TreeMap_prRotateLeft(*entry.TreeMapEntryBase, *rootPointer.Integer)
  Protected.TreeMapEntryBase *newparent
  
  If (*entry <> #Null)
    
    *newparent = *entry\right
    *entry\right = *newparent\left
    
    If (*entry\right <> #Null)
      *entry\right\parent = *entry
    EndIf
    
    *newparent\parent = *entry\parent
    
    If (*entry\parent = #Null)
      *rootPointer\i = *newparent
    ElseIf (*entry = *entry\parent\left)
      *newparent\parent\left = *newparent
    Else
      *newparent\parent\right = *newparent
    EndIf
    
    *newparent\left = *entry
    *entry\parent = *newparent
  EndIf
EndProcedure


;- TreeMap_prRotateRight
Procedure TreeMap_prRotateRight(*entry.TreeMapEntryBase, *rootPointer.Integer)
  Protected.TreeMapEntryBase *newparent
  
  If (*entry <> #Null)
    *newparent = *entry\left
    *entry\left = *newparent\right
    
    If (*entry\left <> #Null)
      *entry\left\parent = *entry
    EndIf
    
    *newparent\parent = *entry\parent
    
    If (*entry\parent = #Null)
      *rootPointer\i = *newparent
    ElseIf (*entry = *entry\parent\left)
      *newparent\parent\left = *newparent
    Else
      *newparent\parent\right = *newparent
    EndIf
    
    *newparent\right = *entry
    *entry\parent = *newparent
  EndIf
EndProcedure


;- TreeMap_prPredecessor
Procedure TreeMap_prPredecessor(*entry.TreeMapEntryBase)
  Protected.TreeMapEntryBase *pre, *last
  
  If (*entry\left <> #Null)
    *pre = *entry\left
    While (*pre\right <> #Null)
      *pre = *pre\right
    Wend
    
  Else
    *last = *entry
    *pre  = *entry\parent
    While (*pre <> #Null) And (*pre\right <> *last)
      *last = *pre
      *pre  = *pre\parent
    Wend
    
  EndIf
  
  ProcedureReturn *pre
EndProcedure


;- TreeMap_prSuccessor
Procedure TreeMap_prSuccessor(*entry.TreeMapEntryBase)
  Protected.TreeMapEntryBase *suc, *last
  
  If (*entry\right <> #Null)
    *suc = *entry\right
    While (*suc\left <> #Null)
      *suc = *suc\left
    Wend
    
  Else
    *last = *entry
    *suc  = *entry\parent
    While (*suc <> #Null) And (*suc\left <> *last)
      *last = *suc
      *suc  = *suc\parent
    Wend
    
  EndIf
  
  ProcedureReturn *suc
EndProcedure


;- TreeMap_prFixDel
Procedure TreeMap_prFixDel(*entry.TreeMapEntryBase, *rootPointer.Integer)
  Protected.TreeMapEntryBase *brother
  Protected *value
  
  While (*entry <> *rootPointer\i) And (*entry\color = #TreeMapColorBlack)
    If (*entry = *entry\parent\left)
      *brother = *entry\parent\right
      
      If (*brother\color = #TreeMapColorRed)
        *brother\color = #TreeMapColorBlack
        *entry\parent\color = #TreeMapColorRed
        TreeMap_prRotateLeft(*entry\parent, *rootPointer)
        *brother = *entry\parent\right
      EndIf
      
      If (TreeMap_prGetColor(*brother\left) = #TreeMapColorBlack) And (TreeMap_prGetColor(*brother\right) = #TreeMapColorBlack)
        *brother\color = #TreeMapColorRed
        *entry = *entry\parent
        
      Else
        If (TreeMap_prGetColor(*brother\right) = #TreeMapColorBlack)
          *brother\left\color = #TreeMapColorBlack
          *brother\color = #TreeMapColorRed
          TreeMap_prRotateRight(*brother, *rootPointer)
          *brother = *entry\parent\right
        EndIf
        
        *brother\color = *entry\parent\color
        *entry\parent\color  = #TreeMapColorBlack
        *brother\right\color = #TreeMapColorBlack
        TreeMap_prRotateLeft(*entry\parent, *rootPointer)
        *entry = *rootPointer\i
        
      EndIf
      
    Else
      *brother = *entry\parent\left
      
      If (*brother\color = #TreeMapColorRed)
        *brother\color = #TreeMapColorBlack
        *entry\parent\color = #TreeMapColorRed
        TreeMap_prRotateRight(*entry\parent, *rootPointer)
        *brother = *entry\parent\left
      EndIf
      
      If (TreeMap_prGetColor(*brother\left) = #TreeMapColorBlack) And (TreeMap_prGetColor(*brother\right) = #TreeMapColorBlack)
        *brother\color = #TreeMapColorRed
        *entry = *entry\parent
        
      Else
        If (TreeMap_prGetColor(*brother\left) = #TreeMapColorBlack)
          *brother\right\color = #TreeMapColorBlack
          *brother\color = #TreeMapColorRed
          TreeMap_prRotateLeft(*brother, *rootPointer)
          *brother = *entry\parent\left
        EndIf
        
        *brother\color = *entry\parent\color
        *entry\parent\color = #TreeMapColorBlack
        *brother\left\color = #TreeMapColorBlack
        TreeMap_prRotateRight(*entry\parent, *rootPointer)
        *entry = *rootPointer\i
        
      EndIf
      
    EndIf
  Wend
  
  *entry\color = #TreeMapColorBlack
EndProcedure


Macro InitTreeMapEntry(_key_, _value_)
  DisableDebugger
  
  CompilerIf Not Defined(TreeMapEntry#_key_#_value_, #PB_Structure)
    
    Structure TreeMapEntry#_key_#_value_ Extends TreeMapEntryBase Align SizeOf(Integer)
      key._key_
      value._value_
    EndStructure
    
    
    ;- TreeMapEntry_getKey
    Procedure TreeMapEntry#_key_#_value_#_getKey(*this.TreeMapEntry#_key_#_value_)
      ProcedureReturn @*this\key
    EndProcedure
    
    
    ;- TreeMapEntry_getValue
    Procedure TreeMapEntry#_key_#_value_#_getValue(*this.TreeMapEntry#_key_#_value_)
      ProcedureReturn @*this\value
    EndProcedure
    
    
    ;- TreeMapEntry_setValue
    Procedure TreeMapEntry#_key_#_value_#_setValue(*this.TreeMapEntry#_key_#_value_, *value)
      CopyStructure(*value, *this\value, _value_)
    EndProcedure
    
    
    ;- NewTreeMapEntry
    Procedure NewTreeMapEntry#_key_#_value_(*this.TreeMapEntry#_key_#_value_ = #Null)
      If (*this = #Null)
        *this = AllocateStructure(TreeMapEntry#_key_#_value_)
      EndIf
      *this\VTable  = ?Functions
      
      DataSection
        Functions:
        Data.i @TreeMapEntry#_key_#_value_#_getKey()
        Data.i @TreeMapEntry#_key_#_value_#_getValue()
        Data.i @TreeMapEntry#_key_#_value_#_setValue()
      EndDataSection
      
      ProcedureReturn *this
    EndProcedure
    
  CompilerEndIf
  
  EnableDebugger
EndMacro


Macro InitTreeMap(_key_, _value_)
  DisableDebugger
  
  InitTreeMapEntry(_key_, _value_)
  InitContainer(TreeMapEntry#_key_#_value_, List)
  
  
  CompilerIf Not Defined(TreeMap#_key_#_value_, #PB_Structure)
    
    Structure TreeMap#_key_#_value_ Extends CNStructure(TreeMapEntry#_key_#_value_, List)
      compare.Comparator
      *root.TreeMapEntry#_key_#_value_
      *current.TreeMapEntry#_key_#_value_
    EndStructure
    
    
    ;- TreeMap_prGetKey
    Procedure TreeMap#_key_#_value_#_prGetKey(*entry.TreeMapEntry#_key_#_value_)
      If (*entry <> #Null)
        ProcedureReturn @*entry\key
      EndIf
      ProcedureReturn #Null
    EndProcedure
    
    
    ;- TreeMap_prGetValue
    Procedure TreeMap#_key_#_value_#_prGetValue(*entry.TreeMapEntry#_key_#_value_)
      If (*entry <> #Null)
        ProcedureReturn @*entry\value
      EndIf
      ProcedureReturn #Null
    EndProcedure
    
    
    ;- TreeMap_prDelEntry
    Procedure TreeMap#_key_#_value_#_prDelEntry(*this.TreeMap#_key_#_value_, *entry.TreeMapEntry#_key_#_value_)
      Protected.TreeMapEntry#_key_#_value_ *rep
      
      *this\current = TreeMap_prPredecessor(*entry)
      
      ; entry has two children
      If (*entry\left <> #Null) And (*entry\right <> #Null)
        ; replace entry with the next element
        *rep = *entry\right
        While (*rep\left <> #Null)
          *rep = *rep\left
        Wend
        
        ; next lines is equal to...
        ;    *entry\key   = *rep\key
        ;    *entry\value = *rep\value
        ;    *entry = *rep
        ; but without copying of key and value
        
        Swap *rep\parent, *entry\parent
        Swap *rep\left,   *entry\left
        Swap *rep\right,  *entry\right
        Swap *rep\color,  *entry\color
        
        *rep\left\parent = *rep
        
        If (*entry\right <> #Null)
          *entry\right\parent = *entry
        EndIf
        
        If (*rep\right = *rep)
          *entry\parent = *rep
          *rep\right    = *entry
          
        Else
          *rep\right\parent = *rep
          
          If (*entry\parent\left = *rep)
            *entry\parent\left = *entry
          Else
            *entry\parent\right = *entry
          EndIf
          
        EndIf
        
        If (*rep\parent = #Null)
          *this\root = *rep
        ElseIf (*rep\parent\left = *entry)
          *rep\parent\left = *rep
        Else
          *rep\parent\right = *rep
        EndIf
        
      EndIf
      
      ; now entry has only at least one child
      If (*entry\left <> #Null)
        *rep = *entry\left
      Else
        *rep = *entry\right
      EndIf
      
      If (*rep <> #Null)
        ; rep is the one child, replace entry with it
        *rep\parent = *entry\parent
        
        If (*entry\parent = #Null)
          *this\root = *rep
        ElseIf (*entry\parent\left = *entry)
          *entry\parent\left = *rep
        Else
          *entry\parent\right = *rep
        EndIf
        
        If (*entry\color = #TreeMapColorBlack)
          TreeMap_prFixDel(*rep, @*this\root)
        EndIf
        
      ElseIf (*entry\parent = #Null)
        ; only one element was in the tree
        *this\root = #Null
        
      Else
        ; entry is a leaf
        If (*entry\color = #TreeMapColorBlack)
          TreeMap_prFixDel(*entry, @*this\root)
        EndIf
        
        If (*entry\parent <> #Null)
          If (*entry = *entry\parent\left)
            *entry\parent\left = #Null
          Else
            *entry\parent\right = #Null
          EndIf
        EndIf
        
      EndIf
      
      ChangeCurrentElement(*this\elements(), *entry)
      DeleteElement(*this\elements())
    EndProcedure
    
    
    ;- TreeMap_prPut
    Procedure TreeMap#_key_#_value_#_prPut(*this.TreeMap#_key_#_value_, *key, *value, onlyIfAbsent.i, *isAdded.Integer)
      Protected.TreeMapEntry#_key_#_value_ *entry, *parent
      Protected cmp.i
      
      *parent = #Null
      *entry = *this\root
      
      ; search the position to put
      While (*entry <> #Null)
        *parent = *entry
        cmp = *this\compare(*key, @*entry\key)
        
        If (cmp < 0)
          *entry = *entry\left
          
        ElseIf (cmp > 0)
          *entry = *entry\right
          
        Else
          ; the key is already contained in the tree
          ; replace the value only if 'onlyIfAbsent' is false
          If (*value <> #Null) And (onlyIfAbsent = #False)
            CopyStructure(*value, *entry\value, _value_)
          EndIf
          
          If (*isAdded <> #Null)
            *isAdded\i = #False
          EndIf
          *this\current = *entry
          
          ProcedureReturn @*entry\value
          
        EndIf
      Wend
      
      ; the key is not yet contained
      ; create new element in the tree
      *entry = AddElement(*this\elements())
      NewTreeMapEntry#_key_#_value_(*entry)
      
      CopyStructure(*key, @*entry\key, _key_)
      If (*value <> #Null)
        CopyStructure(*value, @*entry\value, _value_)
      EndIf
      
      If (*isAdded <> #Null)
        *isAdded\i = #True
      EndIf
      *this\current = *entry
      
      *entry\color = #TreeMapColorRed
      *entry\parent = *parent
      
      If (cmp < 0)
        *parent\left = *entry
      ElseIf (cmp > 0)
        *parent\right = *entry
      Else
        *this\root = *entry
      EndIf
      
      ; fix red-black after insert
      While (TreeMap_prGetParent(*entry) <> #Null) And (*entry\parent\color = #TreeMapColorRed)
        *parent = *entry\parent
        
        If (*parent\parent\left = *parent)
          If (TreeMap_prGetColor(*parent\parent\right) = #TreeMapColorRed)
            *parent\parent\left\color  = #TreeMapColorBlack
            *parent\parent\right\color = #TreeMapColorBlack
            *parent\parent\color = #TreeMapColorRed
            *entry = *parent\parent
            
          Else
            If (*parent\right = *entry)
              TreeMap_prRotateLeft(*parent, @*this\root)
              Swap *parent, *entry
            EndIf
            
            TreeMap_prRotateRight(*parent\parent, @*this\root)
            *parent\color = #TreeMapColorBlack
            *parent\right\color = #TreeMapColorRed
            
          EndIf
          
        Else
          If (TreeMap_prGetColor(*parent\parent\left) = #TreeMapColorRed)
            *parent\parent\left\color  = #TreeMapColorBlack
            *parent\parent\right\color = #TreeMapColorBlack
            *parent\parent\color = #TreeMapColorRed
            *entry = *parent\parent
            
          Else
            If (*parent\left = *entry)
              TreeMap_prRotateRight(*parent, @*this\root)
              Swap *parent, *entry
            EndIf
            
            TreeMap_prRotateLeft(*parent\parent, @*this\root)
            *parent\color = #TreeMapColorBlack
            *parent\left\color = #TreeMapColorRed
            
          EndIf
          
        EndIf
      Wend
      
      *this\root\color = #TreeMapColorBlack
      
      ProcedureReturn @*this\current\value
    EndProcedure
    
    
    ;- TreeMap_size
    Procedure TreeMap#_key_#_value_#_size(*this.TreeMap#_key_#_value_)
      ProcedureReturn ListSize(*this\elements())
    EndProcedure
    
    
    ;- TreeMap_isEmpty
    Procedure TreeMap#_key_#_value_#_isEmpty(*this.TreeMap#_key_#_value_)
      ProcedureReturn Bool(TreeMap#_key_#_value_#_size(*this) = 0)
    EndProcedure
    
    
    ;- TreeMap_put
    Procedure TreeMap#_key_#_value_#_put(*this.TreeMap#_key_#_value_, *key, *value=#Null, *isAdded.Integer=#Null)
      ProcedureReturn TreeMap#_key_#_value_#_prPut(*this, *key, *value, #False, *isAdded)
    EndProcedure
    
    
    ;- TreeMap_putIfAbsent
    Procedure TreeMap#_key_#_value_#_putIfAbsent(*this.TreeMap#_key_#_value_, *key, *value=#Null, *isAdded.Integer=#Null)
      ProcedureReturn TreeMap#_key_#_value_#_prPut(*this, *key, *value, #True, *isAdded)
    EndProcedure
    
    
    ;- TreeMap_getEntry
    Procedure TreeMap#_key_#_value_#_getEntry(*this.TreeMap#_key_#_value_, *key)
      Protected.TreeMapEntry#_key_#_value_ *entry
      Protected cmp.i
      
      *entry = *this\root
      While (*entry <> #Null)
        cmp = *this\compare(*key, @*entry\key)
        
        If (cmp < 0)
          *entry = *entry\left
          
        ElseIf (cmp > 0)
          *entry = *entry\right
          
        Else
          *this\current = *entry
          ProcedureReturn *entry
          
        EndIf
      Wend
      
      ProcedureReturn #Null
    EndProcedure
    
    
    ;- TreeMap_getKey
    Procedure TreeMap#_key_#_value_#_getKey(*this.TreeMap#_key_#_value_, *key)
      Protected.TreeMapEntry#_key_#_value_ *entry
      
      *entry = TreeMap#_key_#_value_#_getEntry(*this, *key)
      ProcedureReturn TreeMap#_key_#_value_#_prGetKey(*entry)
    EndProcedure
    
    
    ;- TreeMap_getValue
    Procedure TreeMap#_key_#_value_#_getValue(*this.TreeMap#_key_#_value_, *key)
      Protected.TreeMapEntry#_key_#_value_ *entry
      
      *entry = TreeMap#_key_#_value_#_getEntry(*this, *key)
      ProcedureReturn TreeMap#_key_#_value_#_prGetValue(*entry)
    EndProcedure
    
    
    ;- TreeMap_getOrDefault
    Procedure TreeMap#_key_#_value_#_getOrDefault(*this.TreeMap#_key_#_value_, *key, *defValue)
      Protected.TreeMapEntry#_key_#_value_ *entry
      
      *entry = TreeMap#_key_#_value_#_getEntry(*this, *key)
      If (*entry <> #Null)
        ProcedureReturn @*entry\value
      EndIf
      
      ProcedureReturn *defValue
    EndProcedure
    
    
    ;- TreeMap_contains
    Procedure TreeMap#_key_#_value_#_contains(*this.TreeMap#_key_#_value_, *key)
      ProcedureReturn Bool(TreeMap#_key_#_value_#_getEntry(*this, *key) <> #Null)
    EndProcedure
    
    
    ;- TreeMap_delete
    Procedure TreeMap#_key_#_value_#_delete(*this.TreeMap#_key_#_value_, *key)
      Protected *entry
      
      *entry = TreeMap#_key_#_value_#_getEntry(*this, *key)
      
      If (*entry <> #Null)
        TreeMap#_key_#_value_#_prDelEntry(*this, *entry)
        ProcedureReturn #True
      EndIf
      
      ProcedureReturn #False
    EndProcedure
    
    
    ;- TreeMap_deleteEntry
    Procedure TreeMap#_key_#_value_#_deleteEntry(*this.TreeMap#_key_#_value_, *entry.TreeMapEntry#_key_#_value_)
      If (*entry <> #Null)
        TreeMap#_key_#_value_#_prDelEntry(*this, *entry)
        ProcedureReturn #True
      EndIf
      
      ProcedureReturn #False
    EndProcedure
    
    
    ;- TreeMap_reset
    Procedure TreeMap#_key_#_value_#_reset(*this.TreeMap#_key_#_value_)
      *this\current = #Null
    EndProcedure
    
    
    ;- TreeMap_prevEntry
    Procedure TreeMap#_key_#_value_#_prevEntry(*this.TreeMap#_key_#_value_, *pointer.Integer=#Null)
      If (*this\current = #Null)
        *this\current = *this\root
        
        If (*this\current <> #Null)
          While (*this\current\right <> #Null)
            *this\current = *this\current\right
          Wend
        EndIf
        
      Else
        *this\current = TreeMap_prPredecessor(*this\current)
        
      EndIf
      
      If (*pointer <> #Null)
        *pointer\i = *this\current
      EndIf
      
      ProcedureReturn *this\current
    EndProcedure
    
    
    ;- TreeMap_nextEntry
    Procedure TreeMap#_key_#_value_#_nextEntry(*this.TreeMap#_key_#_value_, *pointer.Integer=#Null)
      If (*this\current = #Null)
        *this\current = *this\root
        
        If (*this\current <> #Null)
          While (*this\current\left <> #Null)
            *this\current = *this\current\left
          Wend
        EndIf
        
      Else
        *this\current = TreeMap_prSuccessor(*this\current)
        
      EndIf
      
      If (*pointer <> #Null)
        *pointer\i = *this\current
      EndIf
      
      ProcedureReturn *this\current
    EndProcedure
    
    
    ;- TreeMap_prevKey
    Procedure TreeMap#_key_#_value_#_prevKey(*this.TreeMap#_key_#_value_, *pointer.Integer=#Null)
      Protected *res
      
      TreeMap#_key_#_value_#_prevEntry(*this)
      *res = TreeMap#_key_#_value_#_prGetKey(*this\current)
      
      If (*pointer <> #Null)
        *pointer\i = *res
      EndIf
      
      ProcedureReturn *res
    EndProcedure
    
    
    ;- TreeMap_nextKey
    Procedure TreeMap#_key_#_value_#_nextKey(*this.TreeMap#_key_#_value_, *pointer.Integer=#Null)
      Protected *res
      
      TreeMap#_key_#_value_#_nextEntry(*this)
      *res = TreeMap#_key_#_value_#_prGetKey(*this\current)
      
      If (*pointer <> #Null)
        *pointer\i = *res
      EndIf
      
      ProcedureReturn *res
    EndProcedure
    
    
    ;- TreeMap_prevValue
    Procedure TreeMap#_key_#_value_#_prevValue(*this.TreeMap#_key_#_value_, *pointer.Integer=#Null)
      Protected *res
      
      TreeMap#_key_#_value_#_prevEntry(*this)
      *res = TreeMap#_key_#_value_#_prGetValue(*this\current)
      
      If (*pointer <> #Null)
        *pointer\i = *res
      EndIf
      
      ProcedureReturn *res
    EndProcedure
    
    
    ;- TreeMap_nextKey
    Procedure TreeMap#_key_#_value_#_nextValue(*this.TreeMap#_key_#_value_, *pointer.Integer=#Null)
      Protected *res
      
      TreeMap#_key_#_value_#_nextEntry(*this)
      *res = TreeMap#_key_#_value_#_prGetValue(*this\current)
      
      If (*pointer <> #Null)
        *pointer\i = *res
      EndIf
      
      ProcedureReturn *res
    EndProcedure
    
    
    ;- TreeMap_getCurrentEntry
    Procedure TreeMap#_key_#_value_#_getCurrentEntry(*this.TreeMap#_key_#_value_)
      ProcedureReturn *this\current
    EndProcedure
    
    
    ;- TreeMap_setCurrentEntry
    Procedure TreeMap#_key_#_value_#_setCurrentEntry(*this.TreeMap#_key_#_value_, *entry)
      *this\current = *entry
    EndProcedure
    
    
    ;- TreeMap_getCurrentKey
    Procedure TreeMap#_key_#_value_#_getCurrentKey(*this.TreeMap#_key_#_value_)
      ProcedureReturn @*this\current\key
    EndProcedure
    
    
    ;- TreeMap_getCurrentValue
    Procedure TreeMap#_key_#_value_#_getCurrentValue(*this.TreeMap#_key_#_value_)
      ProcedureReturn @*this\current\value
    EndProcedure
    
    
    ;- TreeMap_clear
    Procedure TreeMap#_key_#_value_#_clear(*this.TreeMap#_key_#_value_)
      ClearList(*this\elements())
    EndProcedure
    
    
    ;- TreeMap_free
    Procedure TreeMap#_key_#_value_#_free(*this.TreeMap#_key_#_value_)
      FreeStructure(*this)
    EndProcedure
    
    
    ;- NewTreeMap
    Procedure NewTreeMap#_key_#_value_(*compare.Comparator, *this.TreeMap#_key_#_value_ = #Null)
      If (*this = #Null)
        *this = AllocateStructure(TreeMap#_key_#_value_)
      EndIf
      *this\VTable  = ?Functions
      *this\compare = *compare
      
      DataSection
        Functions:
        Data.i @TreeMap#_key_#_value_#_size()
        Data.i @TreeMap#_key_#_value_#_isEmpty()
        Data.i @TreeMap#_key_#_value_#_put()
        Data.i @TreeMap#_key_#_value_#_putIfAbsent()
        Data.i @TreeMap#_key_#_value_#_getEntry()
        Data.i @TreeMap#_key_#_value_#_getKey()
        Data.i @TreeMap#_key_#_value_#_getValue()
        Data.i @TreeMap#_key_#_value_#_getOrDefault()
        Data.i @TreeMap#_key_#_value_#_contains()
        Data.i @TreeMap#_key_#_value_#_delete()
        Data.i @TreeMap#_key_#_value_#_deleteEntry()
        Data.i @TreeMap#_key_#_value_#_reset()
        Data.i @TreeMap#_key_#_value_#_prevEntry()
        Data.i @TreeMap#_key_#_value_#_nextEntry()
        Data.i @TreeMap#_key_#_value_#_prevKey()
        Data.i @TreeMap#_key_#_value_#_nextKey()
        Data.i @TreeMap#_key_#_value_#_prevValue()
        Data.i @TreeMap#_key_#_value_#_nextValue()
        Data.i @TreeMap#_key_#_value_#_getCurrentEntry()
        Data.i @TreeMap#_key_#_value_#_setCurrentEntry()
        Data.i @TreeMap#_key_#_value_#_getCurrentKey()
        Data.i @TreeMap#_key_#_value_#_getCurrentValue()
        Data.i @TreeMap#_key_#_value_#_clear()
        Data.i @TreeMap#_key_#_value_#_free()
      EndDataSection
      
      ProcedureReturn *this
    EndProcedure
    
  CompilerEndIf
  
  EnableDebugger
EndMacro


Macro NewTreeMap(_key_, _value_, _compare_, _this_ = #Null)
  NewTreeMap#_key_#_value_(_compare_, _this_)
EndMacro


; IDE Options = PureBasic 5.40 LTS (Windows - x86)
; Folding = --
; EnableXP